/*******************************************************************************
*									       *
* HbookClient -- Histo-Scope module called by HistoClient.c to deal with HBOOK *
*		 histograms and ntuples	to be displayed with Histo-Scope.      *
*									       *
*     This module contains the following user-callable routines:	       *
*									       *
*     For HBOOK users:						               *
*									       *
*	 hs_hbook_setup       - Sets up all HBOOK histograms and ntuples in a  *
*			        top directory for use with Histo-Scope.	       *
*									       *
*	 hs_reset_hbook_setup - Call this routine if you have previously       *
*			        called hs_hbook_setup and have booked new      *
*			        histograms or ntuples, or deleted, renamed,    *
*			        rebinned, or resetted existing ones.	       *
*									       *
* Copyright (c) 1992, 1993 Universities Research Association, Inc.	       *
* All rights reserved.							       *
* 									       *
* This material resulted from work developed under a Government Contract and   *
* is subject to the following license:  The Government retains a paid-up,      *
* nonexclusive, irrevocable worldwide license to reproduce, prepare derivative *
* works, perform publicly and display publicly by or for the Government,       *
* including the right to distribute to other Government contractors.  Neither  *
* the United States nor the United States Department of Energy, nor any of     *
* their employees, makes any warranty, express or implied, or assumes any      *
* legal liability or responsibility for the accuracy, completeness, or         *
* usefulness of any information, apparatus, product, or process disclosed, or  *
* represents that its use would not infringe privately owned rights.           *
*                                        				       *
* Fermilab Nirvana GUI Library						       *
* September 14, 1992							       *
*									       *
* Written by Joy Kyriakopulos						       *
*									       *
* Modification History:							       *
*									       *
*									       *
*******************************************************************************/
#include <stdio.h>
#include <string.h>
#ifdef XDR_IBM_MALLOC_BUG
#define malloc mmalloc
#endif
#ifdef VMS	/* ask rpc/types to refrain from declaring malloc */
#define DONT_DECLARE_MALLOC
#endif /*VMS*/  
#include <rpc/types.h>
#ifdef XDR_IBM_MALLOC_BUG
#undef malloc
#define mem_alloc(bsize)	malloc(bsize)
#endif
#include <stdlib.h>
#include <errno.h>
#include <rpc/xdr.h>
#define XDR_FILTERS_HERE 1
#include "histoscope.h"
#include "../histo_util/histo.h"
#include "../histo_util/xdrHisto.h"
#include "../histo_util/histoUtil.h"
#include "HistoClient.h"
#include "HistoClientHB.h"
#include "HbookClient.h"
#include "clientHBList.h"

static histoListHeader *HbookList = NULL;

void histo_hbook_setup(char *topDirectory)
/*
* hs_hbook_setup() sets up all HBOOK histograms and ntuples in the specified
* top directory for use with Histo-Scope.  This routine should be called by
* the user program after booking the histograms and ntuples with HBOOK, and
* also after calling hs_initialize().  If any of these histograms or ntuples
* are deleted, resetted, or re-binned, or if additional ones are created, 
* hs_reset_hbook_setup should be called to reset Histo-Scope's list for the
* top directory.  Unpredictable results may occur if this routine is called
* without first calling the appropriate HBOOK routines.
*/

{
    if (IsHBInit() == HS_NOT_INIT) {
    	fprintf(stderr,
	 "Error calling hs_hbook_setup: hs_initialize must be called first.\n");
	return;
    }
    InitHB();					/* set HB initialized */

    /* get new list for this HBOOK directory */
    GetHbookList(topDirectory);
    
    /* send end-of-list msg to Histo-Scopes */
    SendHBEndL();
}

void histo_reset_hbook_setup(char *topDirectory)
/*
* hs_reset_hbook_setup() should be called by the user after deleting, resetting,
* or re-binning any histograms or ntuples in a top directory used in a
* previous hs_hbook_setup call, or if additional histograms or ntuples are
* created in that top directory.  hs_reset_hbook_setup will reset Histo-Scope's
* list for the specified top directory.
*/

{
    histoListHeader *h, *hNext = NULL;
    int init;

    if ((init = IsHBInit()) != HB_INIT) {
    	if (init == HS_NOT_INIT)
    	    fprintf(stderr, "Error calling hs_reset_hbook_setup:\n\
                hs_initialize and hs_hbook_setup must be called first.\n");
        else
            fprintf(stderr, "Error calling hs_reset_hbook_setup:\n\
           hs_hbook_setup must be called first.\n");
	return;
    }
    
    /* delete all items in top directory from HbookList */
    BeginItemList();
    for (h = HbookList; h != NULL; h = hNext) {
	hNext = h->next;
	if (strncmp(h->item->category, topDirectory+2, strlen(topDirectory)-2)
		== 0) {
	    DelHBListItem(h->item->id, (hsGeneral *)h->item); /* tell HS */
	    DeleteItemFromList(h->item->id, &HbookList);
        }
    }
    EndItemList();
    /* get new list for this HBOOK directory */
    GetHbookList(topDirectory);	    /* this routine calls AddHBItemToClient() */
    					    /* for each item in top directory */
    /* send end-of-list msg to Histo-Scopes */
    SendHBEndL();
}

void AddHBItemToClient(hsGeneral *hsG)
{
    AddItemToList(hsG, &HbookList);
    hsG->id = IncNumOfItems();
    hsG->uid = 0;
    SendHBItem(hsG);
}

hsGeneral *GetHBItemToList(int firstNext)
{
    static histoListHeader *h;
    
    if (firstNext == 0)
    	h = HbookList;
    else 
        h = h->next;
    if (h == NULL)
        return NULL;
    else
        return h->item;
}

hsGeneral *GetHBItemByID(int item_id)
{
    return GetItemByID(item_id, HbookList);
}

void FreeEntireHBList(void)
{
    FreeEntireList(&HbookList);	  /* free HBOOK item list and the items in it */
}

void FreeHBData(hsGeneral *hsG, int mode_ntuple)
{
    hsNTuple *hnt;
    float *first_row, *fft;
    int i;
    if (hsG->hbookID != 0) {
	switch (hsG->type) {
	    case HS_1D_HISTOGRAM:
	    	if (((hs1DHist *)hsG)->bins != NULL) {
	    	    free(((hs1DHist *)hsG)->bins);
	    	    ((hs1DHist *)hsG)->bins = NULL;
	    	}
	    	if (((hs1DHist *)hsG)->pErrs != NULL) {
	    	    free(((hs1DHist *)hsG)->pErrs);
	    	    ((hs1DHist *)hsG)->pErrs = NULL;
	    	}
		break;
	    case HS_2D_HISTOGRAM:
	    	if (((hs2DHist *)hsG)->bins != NULL) {
	    	    free(((hs2DHist *)hsG)->bins);
	    	    ((hs2DHist *)hsG)->bins = NULL;
	    	}
	    	if (((hs2DHist *)hsG)->pErrs != NULL) {
	    	    free(((hs2DHist *)hsG)->pErrs);
	    	    ((hs2DHist *)hsG)->pErrs = NULL;
	    	}
		break;
	    case HS_NTUPLE:
	    	if (((hsNTuple *)hsG)->data != NULL) {
	            hnt = (hsNTuple *) hsG;
	    	    if (mode_ntuple != 0) {
		      first_row = (float *)
		                  malloc(sizeof(float)*hnt->nVariables);
		      fft = first_row;
	    	      for (i=0; i<hnt->nVariables; i++, fft++) 
	    	                  *fft = hnt->data[i];
	    	      free(hnt->data);
	    	      hnt->data = first_row;
	    	    } else {
	    	      free(((hsNTuple *)hsG)->data);
	    	      hnt->data = NULL;
	    	    }
	    	}		  /* GetHbookData uses only data area */
	    	break;
	    default:
	    	fprintf(stderr,
	    	    "hs_update: Found HBOOK item of unknown type.\n");
	}
    }
}

