/*******************************************************************************
*									       *
* fortranBinding.c - FORTRAN interface for Histo-Scope client routines         *
*									       *
*		     This file contains the following functions:	       *
*									       *
*     For all users:						               *
*									       *
*	 hs_initialize       - Initialize the Histo-Scope connection software  *
*			       and set up a potential connection to a Histo-   *
*			       Scope process.				       *
*	 hs_update	     - Updates the Histo-Scope display.		       *
*	 hs_complete	     - Closes all connections w/Histo-Scope processes. *
*	 hs_complete_and_wait- Waits until all Histo-Scopes finish scoping and *
*			       then performs an hs_complete.		       *
*	 hs_histoscope	     - Invokes Histo-Scope as a sub-process.           *
*	 hs_histo_with_config- Invokes Histo-Scope with a configuration file.  *
*	 hs_num_connected_scopes - Returns the number of connected scopes.     *
*									       *
*     For Histo-Scope item users:				               *
*									       *
*	 hs_create_1d_hist   - Books a one-dimensional histogram.	       *
*	 hs_create_2d_hist   - Books a two-dimensional histogram.	       *
*	 hs_create_ntuple    - Defines an n-tuple.  N-tuples have a specified  *
*			       number of variables and automatic storage       *
*			       allocated as they grow.			       *
*	 hs_create_indicator - Creates an indicator (a scalar value).          *
*	 hs_create_control   - Creates a control (a scalar value set by HS).   *
*	 hs_create_trigger   - Creates a trigger (an event set by HS).	       *
*	 hs_create_group     - Creates a group of data items.		       *
*	 hs_fill_1d_hist     - Adds a value to a one-dimensional histogram.    *
*	 hs_fill_2d_hist     - Adds a value to a two-dimensional histogram.    *
*	 hs_fill_ntuple      - Adds an array of real values to an n-tuple.     *
*	 hs_set_indicator    - Sets the value of an indicator.		       *
*	 hs_read_control     - Reads a control (a scalar value set by HS).     *
*	 hs_check_trigger    - Checks whether a trigger has been set by HS.    *
*	 hs_set_1d_errors    - Copies a vector of real numbers as error info   *
*	 hs_set_2d_errors    - Copies an array of real numbers as error info   *
*	 hs_reset	     - Resets all of the bins of a histogram to 0, or  *
*			       removes all of the data from an n-tuple, or     *
*			       sets an indicator to 0.			       *
*	 hs_save_file	     - Saves all current histograms, n-tuples, and     *
*			       indicators in a Histo-Scope-format file.        *
*	 hs_delete	     - Deletes a histogram, n-tuple, or indicator.     *
*	 hs_delete_items     - Deletes a list of histograms, n-tuples, etc.    *
*									       *
*     For HBOOK users:						               *
*									       *
*	 hs_hbook_setup       - Sets up all HBOOK histograms and ntuples in a  *
*			        top directory for use with Histo-Scope.	       *
*	 hs_reset_hbook_setup - Call this routine if you have previously       *
*			        called hs_hbook_setup and have booked new      *
*			        histograms or ntuples, or deleted, renamed,    *
*			        rebinned, or resetted existing ones.	       *
*									       *
* Copyright (c) 1992, 1993 Universities Research Association, Inc.	       *
* All rights reserved.							       *
* 									       *
* This material resulted from work developed under a Government Contract and   *
* is subject to the following license:  The Government retains a paid-up,      *
* nonexclusive, irrevocable worldwide license to reproduce, prepare derivative *
* works, perform publicly and display publicly by or for the Government,       *
* including the right to distribute to other Government contractors.  Neither  *
* the United States nor the United States Department of Energy, nor any of     *
* their employees, makes any warranty, express or implied, or assumes any      *
* legal liability or responsibility for the accuracy, completeness, or         *
* usefulness of any information, apparatus, product, or process disclosed, or  *
* represents that its use would not infringe privately owned rights.           *
*                                        				       *
* Fermilab Nirvana GUI Library						       *
* October 30, 1992							       *
*									       *
* Written by Joy Kyriakopulos						       *
*									       *
* Modified for VMS Port 2/24/93 by Joy Kyriakopulos			       * 
*          for adding hs_create_control and hs_read_control 8/24/93 by JMK     *
*	P. Lebrun, December 10 1993, adding uid arguments and new API routines *
*									       *
*******************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef VMS
#include <sys/descrip.h>
#endif /*VMS*/
#include "../histo_util/histo.h"
#include "../histo_util/histoUtil.h"
#include "HistoClient.h"
#include "histoApiFiles.h"
#ifdef LINKED_WITH_HBOOK
#include "HbookClient.h"
#endif /*LINKED_WITH_HBOOK*/

char *mallocNCopy(char *string, int length);
static void cleanFortranString(char *string, int length);
static void get_config_string(char *cfgFile, char **cfgBuf, int *strLen);

#ifdef VMS
static char **strDescArToNul(struct dsc$descriptor_a *vmsStrAr, int maxDim);
static void FreeArPtrNul(char **strPtrAr);


void hs_initialize(struct dsc$descriptor_s *vmsIdString)
{
    int length = vmsIdString->dsc$w_length;
    char *id_string = vmsIdString->dsc$a_pointer;
#else
void hs_initialize_(char *id_string, int length)
{
#endif /*VMS*/
    char *idString;
    
    idString = mallocNCopy(id_string, length);
    histo_initialize(idString);
    free(idString);
}

#ifdef VMS
void hs_update(void)
#else
void hs_update_(void)
#endif /*VMS*/
{
    histo_update();
}

#ifdef VMS
void hs_complete(void)
#else
void hs_complete_(void)
#endif /*VMS*/
{
    histo_complete();
}

#ifdef VMS
void hs_complete_and_wait(void)
#else
void hs_complete_and_wait_(void)
#endif /*VMS*/
{
    histo_complete_and_wait();
}

#ifdef VMS
void hs_histoscope(int *return_immediately)
#else
void hs_histoscope_(int *return_immediately)
#endif /*VMS*/
{
    histo_histoscope(*return_immediately, NULL, 0);
}
#ifdef VMS
void hs_histo_with_config(int *return_immediately, 
	struct dsc$descriptor_s *config_fileV)
{
    int config_length    = config_fileV->dsc$w_length;
#else
void hs_histo_with_config_(int *return_immediately, char *config_file, 
	int config_length)
{
#endif /*VMS*/
    char *config_fileS = mallocNCopy(config_file, config_length);
    
    histo_histoscope(*return_immediately, config_fileS, 0);
    free(config_fileS);
}

#ifdef VMS
void hs_kill_histoscope(void)
#else
void hs_kill_histoscope_(void)
#endif /*VMS*/
{
    histo_kill_histoscope();
}

#ifdef VMS
int hs_num_connected_scopes(void)
#else
int hs_num_connected_scopes_(void)
#endif /*VMS*/
{
    return histo_num_connected_scopes();
}

#ifdef VMS
void hs_load_config_string(struct dsc$descriptor_s *config_strV)
{
    int cfgBufLen = config_strV->dsc$w_length;
    char *cfgStr = config_strV->dsc$a_pointer;
#else
void hs_load_config_string_(char *cfgStr, int cfgBufLen)
{
#endif /*VMS*/
    int len;
    
    len = strlen(cfgStr);
    histo_load_config(cfgStr, cfgBufLen < len ? cfgBufLen : len);
}

#ifdef VMS
void hs_load_config_file(struct dsc$descriptor_s *config_fileV)
{
    int cfgFLen = config_fileV->dsc$w_length;)
    char *cfgFile = config_fileV->dsc$a_pointer;
#else
void hs_load_config_file_(char *cfgFile, int cfgFLen)
{
    char *cfgFileS;
#endif /*VMS*/
    char *cfgBuf = NULL;
    int cfgBufLen = 0;

    cfgFileS = mallocNCopy(cfgFile, cfgFLen);
    cleanFortranString(cfgFileS, cfgFLen);
    get_config_string(cfgFileS, &cfgBuf, &cfgBufLen);
    if (cfgBuf != 0) {
	histo_load_config(cfgBuf, cfgBufLen);
	free(cfgBuf);
    }
    free(cfgFileS);
}

#ifdef VMS
int hs_create_1d_hist(int *uid, struct dsc$descriptor_s *titleV, 
			struct dsc$descriptor_s *categoryV,  
			struct dsc$descriptor_s *x_labelV, 
			struct dsc$descriptor_s *y_labelV,
			int *n_bins, float *min, float *max)
{
    int title_length    = titleV->dsc$w_length;
    int category_length = categoryV->dsc$w_length;
    int x_label_length  = x_labelV->dsc$w_length;
    int y_label_length  = y_labelV->dsc$w_length;
    char *title    = titleV->dsc$a_pointer;
    char *category = categoryV->dsc$a_pointer;
    char *x_label  = x_labelV->dsc$a_pointer;
    char *y_label  = y_labelV->dsc$a_pointer;
#else
int hs_create_1d_hist_(int *uid, char *title, char *category, char *x_label,
	   char *y_label, int *n_bins, float *min, float *max, int title_length,
	   int category_length, int x_label_length, int y_label_length)
{
#endif /*VMS*/
    char *titleS, *categoryS, *xLabel, *yLabel;
    int hs;
    
    titleS = mallocNCopy(title, title_length);
    categoryS = mallocNCopy(category, category_length);
    xLabel = mallocNCopy(x_label, x_label_length);
    yLabel = mallocNCopy(y_label, y_label_length);
    hs = histo_create_1d_hist(*uid, titleS, categoryS, xLabel, 
    					yLabel, *n_bins, *min, *max);
    free(titleS);
    free(categoryS);
    free(xLabel);
    free(yLabel);
    return hs;
}

#ifdef VMS
int hs_create_2d_hist(int *uid, struct dsc$descriptor_s *titleV, 
			struct dsc$descriptor_s *categoryV,  
			struct dsc$descriptor_s *x_labelV, 
			struct dsc$descriptor_s *y_labelV,
			struct dsc$descriptor_s *z_labelV,
			int *x_bins, int *y_bins, float *x_min, float *x_max,
			float *y_min, float *y_max)
{
    int title_length    = titleV->dsc$w_length;
    int category_length = categoryV->dsc$w_length;
    int x_label_length  = x_labelV->dsc$w_length;
    int y_label_length  = y_labelV->dsc$w_length;
    int z_label_length  = z_labelV->dsc$w_length;
    char *title    = titleV->dsc$a_pointer;
    char *category = categoryV->dsc$a_pointer;
    char *x_label  = x_labelV->dsc$a_pointer;
    char *y_label  = y_labelV->dsc$a_pointer;
    char *z_label  = z_labelV->dsc$a_pointer;
#else
int hs_create_2d_hist_(int *uid, char *title, char *category, 
			char *x_label,char *y_label,
			char *z_label, int *x_bins, int *y_bins, float *x_min,
			float *x_max, float *y_min, float *y_max, int title_length,
			int category_length, int x_label_length, int y_label_length,
			int z_label_length)
{
#endif /*VMS*/
    char *titleS, *categoryS, *xLabel, *yLabel, *zLabel;
    int hs;
    
    titleS = mallocNCopy(title, title_length);
    categoryS = mallocNCopy(category, category_length);
    xLabel = mallocNCopy(x_label, x_label_length);
    yLabel = mallocNCopy(y_label, y_label_length);
    zLabel = mallocNCopy(z_label, z_label_length);
    hs = histo_create_2d_hist(*uid, titleS, categoryS, xLabel, yLabel,
    			 zLabel, *x_bins, *y_bins, *x_min, *x_max,
    			  *y_min, *y_max);
    free(titleS);
    free(categoryS);
    free(xLabel);
    free(yLabel);
    free(zLabel);
    return hs;
}

#ifdef VMS
int hs_create_ntuple(int *uid, struct dsc$descriptor_s *titleV, 
			struct dsc$descriptor_s *categoryV, 
			int *n_variables, struct dsc$descriptor_a *vmsNames)
{
    int title_length    = titleV->dsc$w_length;
    int category_length = categoryV->dsc$w_length;
    char *title    = titleV->dsc$a_pointer;
    char *category = categoryV->dsc$a_pointer;
#else
int hs_create_ntuple_(int *uid, char *title, char *category, int *n_variables,
			char *ftnNames, int title_length, int category_length,
			int name_length)
{
#endif /*VMS*/
    char *titleS, *categoryS;
    char **names;
    int i, nt;
    
    titleS = mallocNCopy(title, title_length);
    categoryS = mallocNCopy(category, category_length);
#ifdef VMS
    names = strDescArToNul(vmsNames, *n_variables);
#else
    names = (char **) malloc(sizeof(char *) * *n_variables);
    for (i = 0; i < *n_variables; ++i)
    	names[i] = mallocNCopy(ftnNames + i*name_length, name_length);
#endif /*VMS*/
    nt = histo_create_ntuple(*uid, titleS, categoryS, *n_variables, names);
    free(titleS);
    free(categoryS);
#ifdef VMS
    FreeArPtrNul(names);
#else
    for (i = 0; i < *n_variables; ++i)
    	free(names[i]);
    free (names);
#endif /*VMS*/
    return (nt);
}

#ifdef VMS
int hs_create_indicator(int *uid, struct dsc$descriptor_s *titleV, 
			struct dsc$descriptor_s *categoryV, 
			float *min, float *max)
{
    int title_length    = titleV->dsc$w_length;
    int category_length = categoryV->dsc$w_length;
    char *title    = titleV->dsc$a_pointer;
    char *category = categoryV->dsc$a_pointer;
#else
int hs_create_indicator_(int *uid, char *title, char *category, 
			float *min, float *max,
			int title_length, int category_length)
{
#endif /*VMS*/
    char *titleS, *categoryS;
    int ind;
    
    titleS = mallocNCopy(title, title_length);
    categoryS = mallocNCopy(category, category_length);
    ind = histo_create_indicator(*uid, titleS, categoryS, *min, *max);
    free(titleS);
    free(categoryS);
    return (ind);
}

#ifdef VMS
int hs_create_control(int *uid, struct dsc$descriptor_s *titleV, 
			struct dsc$descriptor_s *categoryV, 
			float *min, float *max, float *default_value)
{
    int title_length    = titleV->dsc$w_length;
    int category_length = categoryV->dsc$w_length;
    char *title    = titleV->dsc$a_pointer;
    char *category = categoryV->dsc$a_pointer;
#else
int hs_create_control_(int *uid, char *title, char *category,
		 float *min, float *max,
		float *default_value, int title_length, int category_length)
{
#endif /*VMS*/
    char *titleS, *categoryS;
    int ctrl;
    
    titleS = mallocNCopy(title, title_length);
    categoryS = mallocNCopy(category, category_length);
    ctrl = histo_create_control(*uid, titleS, categoryS, *min, 
    				*max, *default_value);
    free(titleS);
    free(categoryS);
    return (ctrl);
}

#ifdef VMS
int hs_create_trigger(int *uid, struct dsc$descriptor_s *titleV, 
			struct dsc$descriptor_s *categoryV)
{
    int title_length    = titleV->dsc$w_length;
    int category_length = categoryV->dsc$w_length;
    char *title    = titleV->dsc$a_pointer;
    char *category = categoryV->dsc$a_pointer;
#else
int hs_create_trigger_(int *uid, char *title, char *category,
			 int title_length, int category_length)
{
#endif /*VMS*/
    char *titleS, *categoryS;
    int trigger;
    
    titleS = mallocNCopy(title, title_length);
    categoryS = mallocNCopy(category, category_length);
    trigger = histo_create_trigger(*uid, titleS, categoryS);
    free(titleS);
    free(categoryS);
    return (trigger);
}

#ifdef VMS
int hs_create_group(int *uid, struct dsc$descriptor_s *titleV, 
			struct dsc$descriptor_s *categoryV, int *groupType,
			int *numItems, int *itemId, int *errsDisp)
{
    int title_length    = titleV->dsc$w_length;
    int category_length = categoryV->dsc$w_length;
    char *title    = titleV->dsc$a_pointer;
    char *category = categoryV->dsc$a_pointer;
#else
int hs_create_group_(int *uid, char *title, char *category, int *groupType,
			 int *numItems, int *itemId, int *errsDisp,
			 int title_length, int category_length)
{
#endif /*VMS*/
    char *titleS, *categoryS;
    int group;
    
    titleS = mallocNCopy(title, title_length);
    categoryS = mallocNCopy(category, category_length);
    group = histo_create_group(*uid, titleS, categoryS, *groupType, *numItems,
    		itemId, errsDisp);
    free(titleS);
    free(categoryS);
    return (group);
}

#ifdef VMS
void hs_fill_1d_hist(int *id, float *x, float *weight)
#else
void hs_fill_1d_hist_(int *id, float *x, float *weight)
#endif /*VMS*/
{
    histo_fill_1d_hist(*id, *x, *weight);
}

#ifdef VMS
void hs_fill_2d_hist(int *id, float *x, float *y, float *weight)
#else
void hs_fill_2d_hist_(int *id, float *x, float *y, float *weight)
#endif /*VMS*/
{
    histo_fill_2d_hist(*id, *x, *y, *weight);
}

#ifdef VMS
int hs_fill_ntuple(int *id, float *values)
#else
int hs_fill_ntuple_(int *id, float *values)
#endif /*VMS*/
{
    return (histo_fill_ntuple(*id, values));
}

#ifdef VMS
void hs_read_control(int *id, float *value)
#else
void hs_read_control_(int *id, float *value)
#endif /*VMS*/
{
    histo_read_control(*id, value);
}

#ifdef VMS
int hs_check_trigger(int *id)
#else
int hs_check_trigger_(int *id)
#endif /*VMS*/
{
    return (histo_check_trigger(*id));
}

#ifdef VMS
void hs_set_indicator(int *id, float *value)
#else
void hs_set_indicator_(int *id, float *value)
#endif /*VMS*/
{
    histo_set_indicator(*id, *value);
}

#ifdef VMS
void hs_set_1d_errors(int *id, float *err_valsP, float *err_valsM)
#else
void hs_set_1d_errors_(int *id, float *err_valsP, float *err_valsM)
#endif /*VMS*/
{
    (histo_set_1d_errors(*id, err_valsP, err_valsM));
}

#ifdef VMS
void hs_set_2d_errors(int *id, float *err_valsP, float *err_valsM)
{
#else
void hs_set_2d_errors_(int *id, float *err_valsP, float *err_valsM)
{
#endif /*VMS*/
    /* Note:  just like HBOOK, we require the array to be dimensioned exactly
    	      as specified to hs_create_2d_hist: x_bins x y_bins. 	      */

    (histo_set_2d_errors(*id, err_valsP, err_valsM, 1));
    						      /* flag col-major order */
}

#ifdef VMS
void hs_reset(int *id)
#else
void hs_reset_(int *id)
#endif /*VMS*/
{
   histo_reset(*id);
}

#ifdef VMS
int hs_save_file(struct dsc$descriptor_s *nameV)
{
    int length = nameV->dsc$w_length;
    char *name = nameV->dsc$a_pointer;
#else
int hs_save_file_(char *name, int length)
{
#endif /*VMS*/
    char *fileName;
    int retStat;
    
    fileName = mallocNCopy(name, length);
    retStat = histo_save_file(fileName);
    free(fileName);
    return (retStat);
}

#ifdef VMS
void hs_delete(int *id)
#else
void hs_delete_(int *id)
#endif /*VMS*/
{
    histo_delete(*id);
}

#ifndef VMS
void hs_delete_items_(int *ids, int *num_ids)
{
    histo_delete_items(ids, *num_ids);
}
#endif /*VMS*/

#ifdef LINKED_WITH_HBOOK
#ifdef VMS
void hs_hbook_setup(struct dsc$descriptor_s *topDirectoryV)
{
    int length         = topDirectoryV->dsc$w_length;
    char *topDirectory = topDirectoryV->dsc$a_pointer;
#else
void hs_hbook_setup_(char *topDirectory, int length)
{
#endif /*VMS*/
    char *topD;
    
    topD = mallocNCopy(topDirectory, length);
    histo_hbook_setup(topD);
    free(topD);
}

#ifdef VMS
void hs_reset_hbook_setup(struct dsc$descriptor_s *topDirectoryV)
{
    int length         = topDirectoryV->dsc$w_length;
    char *topDirectory = topDirectoryV->dsc$a_pointer;
#else
void hs_reset_hbook_setup_(char *topDirectory, int length)
{
#endif /*VMS*/
    char *topD;
    
    topD = mallocNCopy(topDirectory, length);
    histo_reset_hbook_setup(topD);
    free(topD);

}

#endif    /* LINKED_WITH_HBOOK */

#ifdef VMS
/*
 *  strDescArToNul - Convert VMS String Array Descriptor to a C-like
 *		     pointer array of null-terminated strings.  At most
 *		     maxDim number of strings and pointers will be created.
 *		     Returns the address of the malloc'd pointer
 *		     array, which should be free'd when done by
 *		     calling freeArPtrNul().
 */
static char **strDescArToNul(struct dsc$descriptor_a *vmsStrAr, int maxDim)
{
    char **strPtrAr;
    int i, numElems;

    if (vmsStrAr->dsc$b_dtype != DSC$K_DTYPE_T || (vmsStrAr->dsc$b_class !=
    	    DSC$K_CLASS_A && vmsStrAr->dsc$b_class != DSC$K_CLASS_NCA))
    	printf("HS_CREATE_NTUPLE: Warning - descriptor class/type = %d/%d\n%s",
    	    vmsStrAr->dsc$b_class, vmsStrAr->dsc$b_dtype,
    	    "               Expecting 1/14 or 10/14\n");
    if (vmsStrAr->dsc$b_scale != 0)
    	printf("HS_CREATE_NTUPLE: Warning - descriptor scale not zero (%d)\n",
    		vmsStrAr->dsc$b_scale); 
    if (vmsStrAr->dsc$b_digits != 0)
    	printf("HS_CREATE_NTUPLE: Warning - descriptor digits not zero (%d)\n",
    		vmsStrAr->dsc$b_digits); 
    /* if (!vmsStrAr->dsc$b_aflags.dsc$v_fl_coeff)
      printf("HS_CREATE_NTUPLE: Warning - descr has no multipliers block\n"); */
    /* if (!vmsStrAr->dsc$b_aflags.dsc$v_fl_bounds)
      printf("HS_CREATE_NTUPLE: Warning - descriptor has no bounds block\n"); */
    if (vmsStrAr->dsc$b_dimct != 1)
    	printf("HS_CREATE_NTUPLE: Warning - descriptor has > 1 dimension\n");
    if (vmsStrAr->dsc$w_length < 1 || vmsStrAr->dsc$w_length > 4096)
    	printf("HS_CREATE_NTUPLE: Warning - descriptor has bad length (%d)\n",
    		vmsStrAr->dsc$w_length);
    	
    numElems = vmsStrAr->dsc$l_arsize / vmsStrAr->dsc$w_length;
    if (numElems >= maxDim)
    	numElems = maxDim;
    else
    	printf(
   "HS_CREATE_NTUPLE: Warning - fewer variable names than number specified.\n");
 
    strPtrAr = malloc(sizeof(char *) * (numElems + 1));
    for (i = 0; i < numElems; ++i) {
    	strPtrAr[i] = malloc(vmsStrAr->dsc$w_length + 1);
	strncpy(strPtrAr[i], vmsStrAr->dsc$a_pointer+(i*vmsStrAr->dsc$w_length),
		vmsStrAr->dsc$w_length);
	strPtrAr[i][vmsStrAr->dsc$w_length] = '\0';
	cleanFortranString(strPtrAr[i],vmsStrAr->dsc$w_length+1);
    }
    strPtrAr[numElems] = NULL;
    return strPtrAr;
}

/*
 *   FreeArPtrNul - Frees pointer array and its null-terminated strings that
 *		    were created by StrDescArToNul().
 */
static void FreeArPtrNul(char **strPtrAr)
{
    int i = 0;
    
    while (strPtrAr[i] != NULL)
    	free (strPtrAr[i++]);		/* free each string */
    free(strPtrAr);			/* free pointers    */
}

#endif /*VMS*/

char *mallocNCopy(char *string, int length)
{
    char *str;
    
    if (length < 0) {
    	fprintf(stderr, 
    	"hs: Error. Character argument has negative length, setting to null\n");
    	length = 0;
    }
    str = malloc(length + 1);
    if (length > 0)
    	strncpy(str, string, length);
    str[length] = '\0';
    /* printf(" string = /%s/, length: %d (before)\n", str, strlen(str)); */
    cleanFortranString(str, length);
    /* printf(" string = /%s/, length: %d (cleaned)\n", str, strlen(str)); */
    return str;
}

/*
** clean junk out of fortran strings that might not have been
** fully initialized or have trailing blanks.  Assumes enough
** room to append a null at the end of the string.
**
** This cleanFortranString continues after detecting a junk character and
** includes a few more valid characters than other cleanFortranString's.
*/
static void cleanFortranString(char *string, int length)
{
    static char validChars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ\
abcdefghijklmnopqrstuvwxyz1234567890/~!@#$%^&*()_+=-`\"\'?><,.\\[]{}:; \t";
    char *inPtr, *outPtr;
    int i;
    
    /* remove junk characters */
    inPtr = outPtr = string;
    for (i=1; i<=length; i++) {
    	if (strchr(validChars, *inPtr))
    	    *outPtr++ = *inPtr++;
    	else
    	    inPtr++;
    }
    
    /* remove trailing spaces */
    for (outPtr--; outPtr>=string; outPtr--)
    	if (*outPtr != ' ' && *outPtr != '\0')
    	    break;
    
    /* add a null to terminate the string */
    ++outPtr;
    *outPtr = '\0';
}

void get_config_string(char *cfgFile, char **cfgBuf, int *strLen)
{
    int fildes, fileLen;
    struct stat statBuf;

    fildes = open(cfgFile, O_RDONLY);
    if (fildes > 0) {
    	if (fstat(fildes, &statBuf) != 0) {
    	    printf("HS_LOAD_CONFIG_FILE - Error getting file status for %s\n",
    	    	cfgFile);
    	    perror("                    - ");
    	    *cfgBuf = 0;
    	    return;
    	}
    	fileLen = statBuf.st_size;
    	*cfgBuf = malloc(fileLen + 1);
    	if (*cfgBuf == 0) {
    	    printf("HS_LOAD_CONFIG_FILE - Config file too large: %s\n",
    	    	cfgFile);
    	    return;
    	}
    	*strLen = read(fildes, *cfgBuf, fileLen);
    	if (*strLen <= 0) {
    	    printf("HS_LOAD_CONFIG_FILE - Error reading configuration file %s\n",
    	    	cfgFile);
    	    perror("                    - ");
    	    *cfgBuf = 0;
    	    return;
    	}
    	close(fildes);
    	(*cfgBuf)[*strLen] = '\0';
    }
    else {
    	printf("HS_LOAD_CONFIG_FILE - Could not open %s\n", cfgFile);
    	perror("                    - ");
    	*cfgBuf = 0;
    }
}
