/*******************************************************************************
*									       *
* readHbookList.c -- Small and minimal interface between a RZ HBOOK file and   *
* the HistoScope Server Process.  Given a UNIX filename, this routine will fill*
* the Histo.h structure, excluding the Data field.                             * 
*									       *
* Copyright (c) 1991, 1993 Universities Research Association, Inc.	       *
* All rights reserved.							       *
* 									       *
* This material resulted from work developed under a Government Contract and   *
* is subject to the following license:  The Government retains a paid-up,      *
* nonexclusive, irrevocable worldwide license to reproduce, prepare derivative *
* works, perform publicly and display publicly by or for the Government,       *
* including the right to distribute to other Government contractors.  Neither  *
* the United States nor the United States Department of Energy, nor any of     *
* their employees, makes any warrenty, express or implied, or assumes any      *
* legal liability or responsibility for the accuracy, completeness, or         *
* usefulness of any information, apparatus, product, or process disclosed, or  *
* represents that its use would not infringe privately owned rights.           *
*                                        				       *
* Fermilab Nirvana GUI Library						       *
* April 27, 1992							       *
*									       *
* Written by Paul Lebrun						       *
*									       *
*******************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "histo.h"
#include "histoUtil.h"
#include "readHbookList.h"

#define MAXL 132

/* The maximum number of column ( e.g., variable) in an "old type" 
** ( Row wise) Ntuple in HBOOK4 is 512. This is valid for HBOOK 4.21,
** available in CERN V93d and V94a of CERNLIB. See HBOOK documentation
** and subroutine HBOOKN, first few statements where the argument NVAR
** is checked 
*/ 
#define MAX_NUM_VAR_HBOOK_NTUPLE 512

#ifdef VMS
#define TMPFILE "SYS$LOGIN:HISTOXXXXXX"
#include "../util/VMSUtils.h"

static char **strDescArToNul(struct dsc$descriptor_a *vmsStrAr, int maxDim);
static struct dsc$descriptor_a *createStrDescAr(int numChars, int numElems);

/* Function Prototypes for utilHbook.f */
extern int openhbookfile(struct dsc$descriptor_s *fname, int *lrec);
#define closehbookfile_ closehbookfile
extern void mkhbooklist(struct dsc$descriptor_s *fname,
                        struct dsc$descriptor_s *dname);

/* Function Prototypes for HBOOK */
extern void hcdir(struct dsc$descriptor_s *dn, struct dsc$descriptor_s *dm);
static void hcdir_(char *dn, char *dm, int l1, int l2);
extern void hgive(int *i, struct dsc$descriptor_s *dt, int *n1, float *mi1, 
		   float *ma1, int *n2, float *mi2, float *ma2, int *j, int*k);
extern void hgiven(int *i, struct dsc$descriptor_s *dt, int *n1, 
		   struct dsc$descriptor_a *tagsStr, float *mi1, float *ma1);
#define hdelet_ hdelet
#define hrin_ hrin
#define hntnew_ hntnew

#else /* not VMS*/
#define TMPFILE "/tmp/histoXXXXXX"

/* Function Prototypes for utilHbook.f */
extern int openhbookfile_(char *fname, int *lrec, int l);
extern void closehbookfile_();
extern void mkhbooklist_(char *fname, char *dname, int l1, int l2);

/* Function Prototypes for HBOOK */
extern void hcdir_(char *dn, char *dc, int l, int l2);
extern void hgive_(int *i, char *dt, int *n1, float *mi1, float *ma1,
	           int *n2, float *mi2, float *ma2, int *j, 
	           int*k, int l);
extern void hgiven_(int *i, char *dt, int *n1,
                      char tags[MAX_NUM_VAR_HBOOK_NTUPLE][8],
		    float *mi1, float *ma1, int l, int l2);
#endif /*VMS*/

extern void hdelet_(int *i);
extern void hrin_(int *i, int *j, int *k);
extern int hntnew_(int *id);

static void cleanFortranString(char *string, int length);
static void getHbookList(char * dname, int *numRead, hsGeneral ***itemTable);
static void trimLeadNtrailWhite(char *string);

extern char *mktemp();

/*
**	Open the RZ file, which contains Histogram Data, and make a 
**	listing of all histograms
*/ 
int ReadHbookList(char* filename, int lrec, int *numRead, 
			hsGeneral ***addrItemTablePtr)
{ 
	char dname[7];
	int stat;
	
#ifdef VMS
	struct dsc$descriptor_s *ds;

	if (openhbookfile((ds = NulStrToDesc(filename)), &lrec) != 0)
	    return 0;
	FreeStrDesc(ds);
#else
	if (openhbookfile_(filename, &lrec, strlen(filename)) != 0)
	    return 0;
#endif /*VMS*/
	strcpy(dname, "//Top");
	getHbookList(dname, numRead, addrItemTablePtr);
	closehbookfile_(); 
	return 1;  
}


static void getHbookList(char* dname, int *numRead, hsGeneral ***itemTable)
{ 
	int i, nh1, nh2, ntp, nc, lc, l;
	int nb1, nb2, ihn;
	int item;
	float min1, min2, max1, max2;
	char tmpName[60];
	char *p, *tmpdir;
	char **pp;
	char line[MAXL], dirname[81], title[81], nottitle[81];
	char c_space[4];
	float rlow[MAX_NUM_VAR_HBOOK_NTUPLE], rhigh[MAX_NUM_VAR_HBOOK_NTUPLE];
	FILE  *fptl;
	histoListHeader *histoListRead = NULL;
	histoListHeader *hL;
	hs1DHist *hh1;
	hs2DHist *hh2;
	hsNTuple *hntp;
#ifdef VMS
        struct dsc$descriptor_a *tagsStr;
        struct dsc$descriptor_s *titleD, *ds, *dirhsn;
        char **tags, tmpN[61], tmpNdir[MAX_NUM_VAR_HBOOK_NTUPLE];
#else
        char tags[MAX_NUM_VAR_HBOOK_NTUPLE][8];
#endif /*VMS*/

/*
**	Set default to the top-directory.
*/
	strcpy(c_space, "   ");
	hcdir_(dname, c_space, strlen(dname), strlen(c_space));
/*
** 	Make a listing of all histograms on a temporary file
*/
#ifdef VMS
        ds = NulStrWrtDesc(tmpN, 60);
        memset(tmpN, ' ', 60);
        tmpN[61] = '\0';
        strcpy(tmpName, TMPFILE);
        mktemp(tmpName);
        strcat(tmpName, ".DAT");
        strncpy(tmpN, tmpName, strlen(tmpName));
        dirhsn = NulStrWrtDesc(tmpNdir, strlen(dname));
        memset(tmpNdir, ' ', strlen(dname));
        strcpy(tmpNdir, dname);
        mkhbooklist(ds, dirhsn);
        FreeStrDesc(ds);
        FreeStrDesc(dirhsn);
#else
	memset(tmpName, 0, 60);
	strcpy(tmpName, TMPFILE);	
	p = mktemp(tmpName);
	strcpy(tmpName, p);
	mkhbooklist_(tmpName, dname, 60, strlen(dname));
#endif /*VMS*/
/*
** 	Scan this file, and store information.
*/
	ntp = nh1 = nh2 = item = 0;
	fptl = fopen(tmpName,"r");
	while (fgets(line, MAXL, fptl) != NULL) {
	 if (sscanf(line," ===> Directory :  %s", dirname) == 1) {
/*
* Make sure we have the complete directory name as a string without padding
* blank or extraneous characters from the FORTRAN printout.  And don't shorten
* the string artificially.
*/
		tmpdir = (char *) strstr(line, "//");
		strcpy(dirname,tmpdir);
		l = strlen(dirname);
		cleanFortranString(dirname,l);
#ifndef VMS
		dirname[l-1] = '\0';
#endif /*VMS*/
	        hcdir_(dirname, c_space, strlen(dirname), strlen(c_space));
	 }
	 if (sscanf(line,"%d (1)  %s", &ihn,  title) == 2) {
/*
**	One dimension histogram
*/
	  hh1 = (hs1DHist *)malloc(sizeof(hs1DHist));
	  nh1++;
	  item++;
	  hh1->id = item;
	  hh1->type = HS_1D_HISTOGRAM;
	  hh1->hbookID = ihn;
	  hh1->uid = 0;
	  l = 0;
	  i = 999999;
	  hrin_(&ihn, &i, &l);
#ifdef VMS
          hgive(&ihn, (titleD = NulStrWrtDesc(title, 80)), &nb1,
                  &min1, &max1, &nb2, &min2, &max2, &nc, &l);
          hh1->title = StrDescToNul(titleD);
          if (strlen(hh1->title) > nc*4)
                hh1->title[nc*4+1] = '\0';
          FreeStrDesc(titleD);
#else
	  hgive_(&ihn, title, &nb1,  
	         &min1, &max1, &nb2, &min2, &max2, &nc, &l, 80);
/*
**		Store the title
*/ 
	  lc = nc*4;
	  hh1->title = (char *) malloc(lc+1);
	  p = hh1->title;
	  for (i=0; i< (nc*4); i++) p[i] = title[i];
	  cleanFortranString(hh1->title, lc);
#endif /*VMS*/
	  /* trim leading & trailing white space */
	  trimLeadNtrailWhite(hh1->title);
	  hdelet_(&ihn);
/*
**		Store the category field
*/
	  if (strcmp(dirname, "//Top") == 0) {
	      hh1->category = (char *) malloc(5);
	      strcpy(hh1->category, "PAWC");
	  }
	  else {
	      hh1->category = (char *) malloc(strlen(dirname)-1);
	      strcpy(hh1->category,dirname+2);
	  }
	  hh1->nBins = nb1;
	  hh1->min = min1;
	  hh1->max = max1;
	  hh1->bins = NULL;
	  hh1->xScaleType = HS_LINEAR;
	  hh1->xScaleBase = 10.;
	  hh1->xLabel = NULL;
	  hh1->yLabel = NULL;
	  hh1->underflow = 0.;
	  hh1->overflow = 0.;
	  hh1->count = 0;
	  hh1->pErrs = NULL;
	  hh1->mErrs = NULL;
	  
	  AddItemToList((hsGeneral *)hh1, &histoListRead);
	}
	 if (sscanf(line,"%d (2)  %s", &ihn,  title) == 2) {
/*
**	Two dimension histograms.
*/
	  hh2 = (hs2DHist *)malloc(sizeof(hs2DHist));
	  nh2++;
	  item++;
	  hh2->id = item;
	  hh2->type = HS_2D_HISTOGRAM;
	  hh2->hbookID = ihn;
	  hh2->uid = 0;
	  l = 0;
	  i = 999999;
	  hrin_(&ihn, &i, &l);
#ifdef VMS
          hgive(&ihn, (titleD = NulStrWrtDesc(title, 80)), &nb1,
                 &min1, &max1, &nb2, &min2, &max2, &nc, &l);
          hh2->title = StrDescToNul(titleD);
          if (strlen(hh2->title)  > nc*4)
                hh2->title[nc*4+1] = '\0';
          FreeStrDesc(titleD);
#else
	  hgive_(&ihn, title, &nb1, &min1, &max1,
	  	 &nb2, &min2, &max2, &nc, &l, 80);
	  lc = nc * 4;
	  hh2->title = (char *) malloc(lc + 1);
	  p = hh2->title;
	  for (i=0; i< (nc*4); i++) p[i] = title[i];		
	  cleanFortranString(hh2->title, lc);
#endif /*VMS*/
	  /* trim leading & trailing white space */
	  trimLeadNtrailWhite(hh2->title);
	  hdelet_(&ihn);
	  if (strcmp(dirname, "//Top") == 0) {
	      hh2->category = (char *) malloc(5);
	      strcpy(hh2->category, "PAWC");
	  }
	  else {
	      hh2->category = (char *) malloc(strlen(dirname)-1);
	      strcpy(hh2->category,dirname+2);
	  }
	  hh2->nXBins = nb1;
	  hh2->xMin = min1;
	  hh2->xMax = max1;
	  hh2->nYBins = nb2;
	  hh2->yMin = min2;
	  hh2->yMax = max2;
	  hh2->bins = NULL;
	  hh2->xScaleType = HS_LINEAR;
	  hh2->yScaleType = HS_LINEAR;
	  hh2->xLabel = NULL;
	  hh2->yLabel = NULL;
	  hh2->zLabel = NULL;
	  hh2->count = 0;
	  hh2->xScaleBase = 10.;
	  hh2->yScaleBase = 10.;
	  hh2->pErrs = NULL;
	  hh2->mErrs = NULL;
	  for (i = 0; i < 3; ++i)
	      for (l = 0; l < 3; ++l)
	      	  hh2->overflow[i][l] = 0.;

	  AddItemToList((hsGeneral *)hh2, &histoListRead);

	 } 
/*
**	NTuples
*/
	 if (sscanf(line,"%d (N)  %s", &ihn,  nottitle) == 2) {
	  l = 0;
	  i = 999999;
	  hrin_(&ihn, &i, &l);
/*
* Check now that this is not a new type of Ntuple, currently unsupported.
*/
	  if (hntnew_(&ihn) != 1) { 
	    hntp = (hsNTuple *)malloc(sizeof(hsNTuple));
	    ntp++;
	    item++;
	    hntp->id = item;
	    hntp->type = HS_NTUPLE;
	    hntp->hbookID = ihn;
	    hntp->uid = 0;
	    hntp->data = NULL;
	    for (i=0; i<4; i++)
	  	hntp->extensions[i] = NULL;
	    l = 0;
	    i = 999999;
	    nb1 = MAX_NUM_VAR_HBOOK_NTUPLE;
#ifdef VMS
            tagsStr = createStrDescAr(8, MAX_NUM_VAR_HBOOK_NTUPLE);
            hgiven(&ihn, (titleD =  NulStrWrtDesc(title, 80)), &nb1, tagsStr,
                rlow, rhigh);
/* see which title is better on VMS */
/*	    printf("ntuple: HCDIRtitle = %s, HGIVENtitle = %s\n", nottitle,title); */
/*          hntp->title = StrDescToNul(titleD);	*/
             tags = strDescArToNul(tagsStr, MAX_NUM_VAR_HBOOK_NTUPLE);
#else
	   hgiven_(&ihn, title, &nb1, tags, rlow, rhigh, 80, 8); 
#endif /*VMS*/
	   hdelet_(&ihn);
	   hntp->nVariables = nb1;
	   hntp->n = 0;
	   hntp->chunkSzData = 0;
	   hntp->chunkSzExt = 0;
	   hntp->data = 0;
	   lc = 80;
	   hntp->title = (char *) malloc(lc + 1);
	   p = hntp->title;
	   for (i=0; i< lc; i++) p[i] = title[i];		
	   cleanFortranString(hntp->title, lc);
	   /* trim leading & trailing white space */
	   trimLeadNtrailWhite(hntp->title);
	   if (strcmp(dirname, "//Top") == 0) {
	       hntp->category = (char *) malloc(5);
	       strcpy(hntp->category, "PAWC");
	   }
	   else {
	       hntp->category = (char *) malloc(strlen(dirname)-1);
	       strcpy(hntp->category,dirname+2);
	   }
#ifdef VMS
          FreeStrDesc(titleD);
          hntp->names = tags;
           for (i=0; i<nb1; ++i) {
          	cleanFortranString(tags[i], strlen(tags[i]));
          	trimLeadNtrailWhite(tags[i]);
           }
#else
	   hntp->names = (char **) malloc(nb1*sizeof(char *));
	   pp = hntp->names;
	   for (i=0; i<nb1; i++) {
	  	*pp = (char *)malloc(9);
	  	memset(*pp, 0, 9);
	  	strncpy(*pp, tags[i], 8);
	  	cleanFortranString(*pp, 8);
	  	trimLeadNtrailWhite(*pp);
	  	pp++;
	   }
#endif /*VMS*/
	   AddItemToList((hsGeneral *)hntp, &histoListRead);
	  } else { /* New type, disregard */
	   hdelet_(&ihn);
	  }
	 } 
	}
	fclose(fptl);
	remove(tmpName);
	
	/* put every item into the item table to return to caller      */
	*numRead = item;		/* return number of items read */
	*itemTable = (hsGeneral **) malloc(sizeof(hsGeneral *) * item);
	for (i = item, hL = histoListRead; hL != NULL; hL = hL->next)
	    (*itemTable)[--i] = hL->item;
	FreeEntireListND(&histoListRead);	/* no longer needed */   
}

/*
** clean junk out of fortran strings that might not have been
** fully initialized or have trailing blanks.  Assumes enough
** room to append a null at the end of the string.
*/
static void cleanFortranString(char *string, int length)
{
    static char validChars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ\
abcdefghijklmnopqrstuvwxyz1234567890/~!@#$%^&*()_+=-`\"\'?><,.\\[]{}:; \t";
/* abcdefghijklmnopqrstuvwxyz1234567890/~!@#$%^&*()_+=-`\"\'?><,. \t"; */
    char *inPtr, *outPtr;
    int i;
    
    /* remove junk characters */
    inPtr = outPtr = string;
    for (i=1; i<=length; i++) {
    	if (strchr(validChars, *inPtr))
    	    *outPtr++ = *inPtr++;
    	else
    	    break;  /* JMK 7/28/93. was: inPtr++;	*/
    }
    
    /* remove trailing spaces */
    for (outPtr--; outPtr>=string; outPtr--)
    	if (*outPtr != ' ')
    	    break;
    
    /* add a null to terminate the string */
    ++outPtr;
    *outPtr = '\0';
}

/*
** trimLeadNtrailWhite - Trim off leading and trailing spaces & tabs
**			 from an input string
** 			 If string is NULL, NULL is returned.
*/
static void trimLeadNtrailWhite(char *string)
{
    char *c;
    int len, whiteLen, i;
    
    if (string == NULL || string[0] == '\0')
    	return;
    
    /* Find the first non-white character & move non-white chars left */
    for (whiteLen = 0; isspace(string[whiteLen]) != 0; ++whiteLen);
     
    for ( c = &string[whiteLen], i = 0, len = strlen(c); 
          whiteLen != 0 && i < len; ++i)
    	string[i] = c[i];
    for ( len = strlen(string); whiteLen != 0 && i < len; ++i)
    	string[i] = ' ';		/* pad with spaces */     	
    
    /* Remove trailing whitespace by converting to nulls */
    if (len == 0)		/* special case for all white space strings */
    	return;
    for (i = len-1; i >= 0; --i) {
    	if (isspace(string[i]))
    	    string[i] = '\0';
    	else
    	    break;
    }
}    

#ifdef VMS

static void hcdir_(char *dn, char *dm, int l1, int l2)
{
    struct dsc$descriptor *ds1, *ds2;

    hcdir((ds1 = NulStrToDesc(dn)), (ds2 = NulStrToDesc(dm)));
    FreeStrDesc(ds1);
    FreeStrDesc(ds2);
}


/*
 *  createStrDescAr - Create VMS String Array Descriptor.
 */
static struct dsc$descriptor_a *createStrDescAr(int numChars, int numElems)
{
    struct dsc$descriptor_a *vmsStrAr;
    struct dsc$descriptor_a_ext {
	struct dsc$descriptor_a descrpt;
	char *dsc$a_a0;			/* multipliers block */
	long dsc$l_m;
	long dsc$l_l;			/* bounds block      */
	long dsc$l_u;
    } *vmsStr;

    vmsStrAr = malloc(sizeof(struct dsc$descriptor_a_ext) + numChars*numElems);
    vmsStrAr->dsc$b_dtype = DSC$K_DTYPE_T;
    vmsStrAr->dsc$b_class = DSC$K_CLASS_A;
    vmsStrAr->dsc$b_scale = 0;
    vmsStrAr->dsc$b_digits = 0;
    vmsStrAr->dsc$b_aflags.dsc$v_fl_coeff = 1;  /* for multipliers block  */
    vmsStrAr->dsc$b_aflags.dsc$v_fl_bounds = 1; /* for bounds block       */
    vmsStrAr->dsc$b_aflags.dsc$v_fl_redim = 0;
    vmsStrAr->dsc$b_aflags.dsc$v_fl_column = 1;
    vmsStrAr->dsc$b_aflags.dsc$v_fl_binscale = 0;
    vmsStrAr->dsc$b_dimct = 1;			/* 1 dimension		  */
    vmsStrAr->dsc$w_length = numChars;
    vmsStrAr->dsc$l_arsize = numChars * numElems;
    vmsStrAr->dsc$b_scale = 0;
    vmsStrAr->dsc$b_digits = 0;
    vmsStrAr->dsc$a_pointer = ((char *)vmsStrAr) + 
				sizeof(struct dsc$descriptor_a_ext);
    vmsStr = (struct dsc$descriptor_a_ext *) vmsStrAr;
    vmsStr->dsc$a_a0 = vmsStrAr->dsc$a_pointer - numChars;
    vmsStr->dsc$l_m = numChars;
    vmsStr->dsc$l_l = 1;
    vmsStr->dsc$l_u = numElems;
 
    return vmsStrAr;
}

/*
 *  strDescArToNul - Convert VMS String Array Descriptor to a C-like
 *		     pointer array of null-terminated strings.  At most
 *		     maxDim number of strings and pointers will be created.
 *		     Returns the address of the malloc'd pointer array.
 */
static char **strDescArToNul(struct dsc$descriptor_a *vmsStrAr, int maxDim)
{
    char **strPtrAr;
    int i, numElems;

    if (vmsStrAr->dsc$b_dtype != DSC$K_DTYPE_T || vmsStrAr->dsc$b_class !=
    	    DSC$K_CLASS_A)
    	printf("HistoScope: Warning - descriptor class/type = %d/%d\n%s",
    	    vmsStrAr->dsc$b_class, vmsStrAr->dsc$b_dtype,
    	    "               Expecting 1/4\n");
    if (vmsStrAr->dsc$b_scale != 0)
    	printf("HistoScope: Warning - descriptor scale not zero (%d)\n",
    		vmsStrAr->dsc$b_scale); 
    if (vmsStrAr->dsc$b_digits != 0)
    	printf("HistoScope: Warning - descriptor digits not zero (%d)\n",
    		vmsStrAr->dsc$b_digits); 
    if (!vmsStrAr->dsc$b_aflags.dsc$v_fl_coeff)
    	printf("HistoScope: Warning - descr has no multipliers block\n");
    if (!vmsStrAr->dsc$b_aflags.dsc$v_fl_bounds)
    	printf("HistoScope: Warning - descriptor has no bounds block\n");
    if (vmsStrAr->dsc$b_dimct != 1)
    	printf("HistoScope: Warning - descriptor has > 1 dimension\n");
    if (vmsStrAr->dsc$w_length < 1 || vmsStrAr->dsc$w_length > 4096)
    	printf("HistoScope: Warning - descriptor has bad length (%d)\n",
    		vmsStrAr->dsc$w_length);
    	
    numElems = vmsStrAr->dsc$l_arsize / vmsStrAr->dsc$w_length;
    if (numElems >= maxDim)
    	numElems = maxDim;
 
    strPtrAr = malloc(sizeof(char *) * (numElems + 1));
    for (i = 0; i < numElems; ++i) {
    	strPtrAr[i] = malloc(vmsStrAr->dsc$w_length + 1);
	strncpy(strPtrAr[i], ((char *)vmsStrAr->dsc$a_pointer) + 
		(i*vmsStrAr->dsc$w_length), vmsStrAr->dsc$w_length);
	strPtrAr[i][vmsStrAr->dsc$w_length] = '\0';
    }
    strPtrAr[numElems+1] = NULL;
    return strPtrAr;
}

#endif /*VMS*/

