/***********************************************************************
*                                                                      *
* cellExample.c - Cell Plot Widget Example Program                     *
*                                                                      *
* This program is an example program that demonstrates the Cell        *
* widget capabilities and how to use it. The program uses a form       *
* widget to contain the widgets that invoke features of the Cell       *
* widget.                                                              *
*                                                                      *
* A rowcolumn widget includes 7 push button widgets, to: set the       *
* visable range of the plot, get the visible range, zoom in,           *
* zoom out, reset zoom, print the plot, and exit this program.         *
*                                                                      *
*                                                                      *
*                                                                      *
* September 9, 1994                                                    *
*                                                                      *
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <limits.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/LabelG.h>
#include <Xm/BulletinB.h>
#include <Xm/Text.h>
#include <Xm/MessageB.h>
#include <Xm/ToggleB.h>
#include "../util/printUtils.h"
#include "Cell.h"
#include "Scat.h"

#define F_RANDOM()    (((double) rand() / ((RAND_MAX + 1.0) / 2.0)) - 1.0)
#define NUM_TEXT   5
#define NUM_X_BINS 35
#define NUM_Y_BINS 35

static Widget Text[NUM_TEXT];
static Widget Cell;

static float *createPoints();
static int nFermiLogo(float x, float y);
static void getAndSetTheCellPoints(Widget Cell, CellRect **ptry2, int *nptry2);
static void createPushButtonsForRowcolumn(Widget rowColumn);
static void setRangeCallback(Widget w, caddr_t client_d, caddr_t call_d);
static void getRangeCallback(Widget w, caddr_t client_d, caddr_t call_d);  
static void okCallback(Widget w, Widget client_d, caddr_t call_d);
static void cancelCallback(Widget w, Widget client_d, caddr_t call_d);
static void zoomCallback(Widget w, XtPointer client_d, caddr_t call_d);
static void exitCallback(Widget w, caddr_t client_d, caddr_t call_d);
static void createWarningMessage(Widget parent, char *text);
static void printCallback(Widget w, caddr_t client_d, caddr_t call_d);


main(unsigned int argc, char *argv[])
{
    XtAppContext app_context;
    Widget toplevel, form, rowColumn;
    XmString axisXLabel, axisYLabel;
    CellRect *ptry2;
    int nptry2;
        

    /* Initialize the Intrinsics. */   
    toplevel = XtAppInitialize(&app_context, "CellTest", NULL,
        0, &argc, argv, NULL, NULL, 0);
    
    /* Create the strings to display the X and Y axis message. */
    axisXLabel = XmStringCreateSimple("This is the X axis label");
    axisYLabel = XmStringCreateSimple("This is the Y axis label");
    

    /* Create a form widget as the container for rowcolumn, 
       Cell, and checkbox. */
    form = XtCreateManagedWidget("form", xmFormWidgetClass, toplevel, NULL, 0);

    /* Create a cell widget to display the points. */
    Cell = XtVaCreateManagedWidget("Cell", cellWidgetClass, form,
        XmNxAxisLabel,       axisXLabel,
        XmNyAxisLabel,       axisYLabel,
        XmNdoubleBuffer,     FALSE, 
        XmNtopAttachment,    XmATTACH_WIDGET,
    	XmNleftAttachment,   XmATTACH_POSITION,
    	XmNleftPosition,     21,
    	XmNrightAttachment,  XmATTACH_WIDGET,
    	XmNbottomAttachment, XmATTACH_WIDGET, NULL); 
    	  
    XmStringFree(axisXLabel);    	
    XmStringFree(axisYLabel);

     
    getAndSetTheCellPoints(Cell, &ptry2, &nptry2); 
    CellSetContents(Cell, ptry2, nptry2, CELL_RESCALE); 
   
   	
    /* Create a rowColumn widget  to hold 7 push buttons. */
    rowColumn = XtVaCreateManagedWidget("rowColumn", xmRowColumnWidgetClass,
    	form, XmNresizable, TRUE,
    	XmNtopAttachment,    XmATTACH_FORM,
    	XmNtopOffset,        10,
    	XmNleftAttachment,   XmATTACH_WIDGET,
    	XmNrightAttachment,  XmATTACH_POSITION,
    	XmNrightPosition,    20,
    	XmNbottomAttachment, XmATTACH_WIDGET, NULL);
    
    createPushButtonsForRowcolumn(rowColumn);

    
    
    /* Display all widgets and call XtMainLoop to continuously process
       all events. */
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_context);
}

static void getAndSetTheCellPoints(Widget Cell, CellRect **ptry2, int *nptry2)
{
    /* This function will initialize the CellRect structure. The
       points will be a two dimensional array. The following 
       algorithm was copied from Mark Edel.  */ 
       
    int i, j, nRects;
    float *bins;
    float xMin = -1.0, yMin = -1.0, xMax = 1.0, yMax = 1.0;
    float xBinWidth, yBinWidth, xScale, yScale, binValue;
    float minBinValue = FLT_MAX, maxBinValue = -FLT_MAX;
    CellRect *rect;

    
    bins = createPoints();
    
    /* find the range of the data: minBinValue & maxBinValue */
    for (i=0; i<NUM_X_BINS; i++) {
    	for (j=0; j<NUM_Y_BINS; j++) {
    	    binValue = bins[i*NUM_Y_BINS + j];
    	    if (binValue < minBinValue) minBinValue = binValue;
    	    if (binValue > maxBinValue) maxBinValue = binValue;
	}
    }
    
   
       
    /* calculate constants used in creating rectangles for cell widget */
    xBinWidth = (xMax - xMin)/NUM_X_BINS;
    yBinWidth = (yMax - yMin)/NUM_Y_BINS;
    xScale = xBinWidth / (maxBinValue - minBinValue);
    yScale = yBinWidth / (maxBinValue - minBinValue);
    nRects = NUM_X_BINS * NUM_Y_BINS;
    
    
    /* create the rectangles for the cell widget to draw */
    nRects = (NUM_X_BINS * NUM_Y_BINS);
    rect = (CellRect *)XtMalloc(nRects * sizeof(CellRect));
    *ptry2 = rect;
    *nptry2 = nRects+1;
    for (i=0; i<NUM_X_BINS; i++) {
    	for (j=0; j<NUM_Y_BINS; j++) {
    	    binValue = bins[i*NUM_Y_BINS + j];
    	    rect->x = xMin + i*xBinWidth + xBinWidth/2;
    	    rect->y = yMin + j*yBinWidth + yBinWidth/2;
    	    rect->dx = (binValue - minBinValue) * xScale;
    	    rect->dy = (binValue - minBinValue) * yScale;
    	    rect->pixel = BlackPixel(XtDisplay(Cell), 0);
    	    rect++;
    	}
    }

}

static float *createPoints()
{
    /* This function will create the points to for the cell.
       The points will be arranged to display the FermiLab logo.
       First, a two dimensional array is created with twice the
       size of NUM_X_BINS by NUM_Y_BINS (bins2) and another two
       dimensional array of NUM_X_BINS by NUM_Y_BINS (bins). Second,
       four points will be averaged and placed in the bins array.
       A pointer to the beginning of the points will be returned. */ 
       
    int i, j, k, l, kk, counter=1;
    float x, y, *logo, *bins, *logo2, *tmpLogo2, *bins2, iDiv, jDiv, total;
    
    
    k = NUM_X_BINS * NUM_Y_BINS;
    kk = k * 4;
    bins = (float *)XtMalloc(sizeof(float) * k);
    bins2 = (float *)XtMalloc(sizeof(float) * kk);
   
    
    /* Initialize to zero. */
    logo2 = bins2;
    for (i = 0; i < (NUM_X_BINS * 2); i++) 
    	for (j = 0; j < (NUM_Y_BINS * 2); j++)
    	    *logo2++ = 0.0; 
   
    
    /* Create Fermi logo points. */	  
    iDiv = (NUM_X_BINS * 2) / 2.0;       /* half of NUM_X_BINS */
    jDiv = (NUM_Y_BINS * 2) / 2.0;       /* half of NUM_Y_BINS */	    	    
    logo2 = bins2;	    
    for (i = 0; i < (NUM_X_BINS * 2); i++)
    	for (j = 0; j < (NUM_Y_BINS * 2); j++) {
    	    x = -1.0 + i / iDiv + 0.005;
    	    y = -1.0 + j / jDiv + 0.005; 
    	    if (nFermiLogo(x,y) ) 
    	        *logo2 = 1.0; 
    	    logo2++;
    	 }


    /* Average the four points and place the average into bins. */
    logo = bins;
    logo2 = bins2;
    tmpLogo2 = bins2;
    for (i = 0; i < NUM_X_BINS; i++) {
    	tmpLogo2 += NUM_Y_BINS * 2;
    	for (j = 0; j < NUM_Y_BINS ; j++) { 
    	    total = 0.0;
    	    for (k = j; k < j+2; k++) 
    	    	total += *logo2++;
    	    for (l = j; l < j+2; l++) 
    	    	total += *tmpLogo2++;
    	    *logo++ = total / 4.0;
    	}
    	logo2 += NUM_X_BINS * 2;
    }
    	
    		
    free(bins2); 
    return (bins);
}


static void createPushButtonsForRowcolumn(Widget rowColumn)
{
    /* Create 7 pushButton widgets for rowColumn. */
    
    Widget setRange, getRange, zoomIn, zoomOut, resetZoom,
           printFile, exitProg;
    XmString xmstr;
    
    
    xmstr = XmStringCreateSimple("Set Visible Range...");
    setRange = XtVaCreateManagedWidget("setRange", xmPushButtonWidgetClass,
    	rowColumn, XmNlabelString, xmstr, NULL);
    XtAddCallback(setRange, XmNactivateCallback, setRangeCallback, NULL);
    XmStringFree(xmstr);
    
    xmstr = XmStringCreateSimple("Get Visible Range...");
    getRange = XtVaCreateManagedWidget("getRange", xmPushButtonWidgetClass,
    	rowColumn, XmNlabelString, xmstr, NULL);
    XtAddCallback(getRange, XmNactivateCallback, getRangeCallback, NULL);
    XmStringFree(xmstr);
    
    xmstr = XmStringCreateSimple("Zoom In");	   	  	
    zoomIn = XtVaCreateManagedWidget("zoomIn", xmPushButtonWidgetClass,
    	rowColumn, XmNlabelString, xmstr, NULL);
    XtAddCallback(zoomIn, XmNactivateCallback, 
    	(XtCallbackProc) zoomCallback, (XtPointer) 1); 
    XmStringFree(xmstr);
    
    xmstr = XmStringCreateSimple("Zoom Out");	   	  	
    zoomOut = XtVaCreateManagedWidget("zoomOut", xmPushButtonWidgetClass,
    	rowColumn, XmNlabelString, xmstr, NULL);
    XtAddCallback(zoomOut, XmNactivateCallback, 
    	(XtCallbackProc) zoomCallback, (XtPointer) 2); 
    XmStringFree(xmstr);
    
    xmstr = XmStringCreateSimple("Reset Zoom");	   	  	
    resetZoom = XtVaCreateManagedWidget("resetZoom", xmPushButtonWidgetClass,
    	rowColumn, XmNlabelString, xmstr, NULL);
    XtAddCallback(resetZoom, XmNactivateCallback, 
    	(XtCallbackProc) zoomCallback, (XtPointer) 3);
    XmStringFree(xmstr);
    
    xmstr = XmStringCreateSimple("Print Cell...");
    printFile = XtVaCreateManagedWidget("printFile", xmPushButtonWidgetClass,
        rowColumn, XmNlabelString, xmstr, NULL);
    XtAddCallback(printFile, XmNactivateCallback, printCallback, NULL);	
    XmStringFree(xmstr);
    
    xmstr = XmStringCreateSimple("Exit Program");  	   	  	
    exitProg = XtVaCreateManagedWidget("exitProg", xmPushButtonWidgetClass,
    	rowColumn, XmNlabelString, xmstr, NULL);
    XtAddCallback(exitProg, XmNactivateCallback, exitCallback, NULL);
    XmStringFree(xmstr);

}

static void setRangeCallback(Widget w, caddr_t client_data, caddr_t call_data)
{
    
    /* This function will create and pop-up a form widget, in which a user
       can enter the maximum and minimum values for the new plot range.
       Also, included is a Cancel and an Ok push button. The callback for
       the Ok pushbutton calls CellSetVisibleRange. */ 
       
    Widget form, labelMinX, labelMinY, labelMaxX, labelMaxY;
    Widget ok, cancel;
    XmString xmstr; 
    
    /* Create a form widget to contain the text & label widgets. */
    xmstr = XmStringCreateSimple("Set Range Dialog");
    form = XmCreateFormDialog(w, "form", NULL, 0);
    XtVaSetValues(form, 
    	XmNdialogTitle, 	xmstr,
    	XmNautoUnmanage,	FALSE, 
    	NULL);
    XmStringFree(xmstr); 
    	
    /* Create 4 label & 4 text widgets. */
    xmstr = XmStringCreateSimple("min X");
    labelMinX = XtVaCreateManagedWidget("labelMinX", 
    	xmLabelGadgetClass, form,
    	XmNlabelString,		xmstr,
    	XmNtopAttachment,	XmATTACH_FORM,
    	XmNtopOffset,		10,
    	XmNleftAttachment,	XmATTACH_FORM,
    	XmNleftOffset,		10,
    	NULL);
    XmStringFree(xmstr); 
    
    Text[0] = XtVaCreateManagedWidget("textMinX", 
    	xmTextWidgetClass, form,
    	XmNtopAttachment,	XmATTACH_FORM,
    	XmNtopOffset,		5,
    	XmNleftAttachment,	XmATTACH_WIDGET,
    	XmNleftWidget,          labelMinX,
    	NULL); 
    	
    xmstr = XmStringCreateSimple("min Y");
    labelMinY = XtVaCreateManagedWidget("labelMinY", 
    	xmLabelGadgetClass, form,
    	XmNlabelString,		xmstr,
    	XmNtopAttachment,	XmATTACH_FORM,
    	XmNtopOffset,		10,
    	XmNleftAttachment,	XmATTACH_WIDGET,
    	XmNleftWidget,          Text[0],
    	XmNleftOffset,		10,
    	NULL);
    XmStringFree(xmstr);
    	
    Text[1] = XtVaCreateManagedWidget("textMinY", 
    	xmTextWidgetClass, form,
    	XmNtopAttachment,	XmATTACH_FORM,
    	XmNtopOffset,		5,
    	XmNleftAttachment,	XmATTACH_WIDGET,
    	XmNleftWidget,          labelMinY,
    	NULL);
    	
    xmstr = XmStringCreateSimple("max X");	
    labelMaxX = XtVaCreateManagedWidget("labelMaxX", 
    	xmLabelGadgetClass, form,
    	XmNlabelString,		xmstr,
    	XmNtopAttachment,	XmATTACH_WIDGET,
    	XmNtopWidget,           labelMinX,
    	XmNtopOffset,		20,
    	XmNleftAttachment,	XmATTACH_FORM,
    	XmNleftOffset,		10,
    	NULL);
    XmStringFree(xmstr);
    
    Text[2] = XtVaCreateManagedWidget("textMaxX", 
    	xmTextWidgetClass, form,
    	XmNtopAttachment,	XmATTACH_WIDGET,
    	XmNtopWidget,           Text[0],
    	XmNtopOffset,		5,
    	XmNleftAttachment,	XmATTACH_WIDGET,
    	XmNleftWidget,          labelMaxX,
    	NULL); 
    	
    xmstr = XmStringCreateSimple("max Y");	
    labelMaxY = XtVaCreateManagedWidget("labelMaxY", 
        xmLabelGadgetClass, form,
        XmNlabelString,		xmstr,
        XmNtopAttachment,	XmATTACH_WIDGET,
    	XmNtopWidget,           labelMinY,
    	XmNtopOffset,		20,
    	XmNleftAttachment,	XmATTACH_WIDGET,
    	XmNleftWidget,          Text[2],
    	XmNleftOffset,		10,
    	NULL);
    XmStringFree(xmstr); 
    
    Text[3] = XtVaCreateManagedWidget("textMaxY", 
    	xmTextWidgetClass, form,
    	XmNtopAttachment,	XmATTACH_WIDGET,
    	XmNtopWidget,           Text[1],
    	XmNtopOffset,		5, 
    	XmNleftAttachment,	XmATTACH_WIDGET,
    	XmNleftWidget,          labelMaxY,
    	NULL);	

    	
    /* Create OK push button widget and callback */
    xmstr = XmStringCreateSimple(" Ok ");
    ok = XtVaCreateManagedWidget("ok", 
    	xmPushButtonWidgetClass, form,
    	XmNlabelString,	 	xmstr,
    	XmNtopAttachment,	XmATTACH_WIDGET,
    	XmNtopWidget,           Text[2],
    	XmNtopOffset,		20,
    	XmNleftAttachment,	XmATTACH_FORM,
    	XmNleftOffset,		20,  
    	NULL);
    XtAddCallback(ok, XmNactivateCallback, 
    	(XtCallbackProc) okCallback, (void *) form);
    XmStringFree(xmstr);
    
    /* Create Cancel push button widget and callback */ 
    xmstr = XmStringCreateSimple("Cancel");  
    cancel = XtVaCreateManagedWidget("cancel", 
    	xmPushButtonWidgetClass, form,
    	XmNlabelString, 	xmstr,
    	XmNtopAttachment,	XmATTACH_WIDGET,
    	XmNtopWidget,           Text[2],
    	XmNtopOffset,		20,
    	XmNleftAttachment,	XmATTACH_WIDGET,
    	XmNleftWidget,          ok,
    	XmNleftOffset,		5, 
    	NULL);
    XtAddCallback(cancel, XmNactivateCallback, 
    	(XtCallbackProc) cancelCallback, (void *) form);
    XmStringFree(xmstr);
    

    XtManageChild(form);
}

static void okCallback(Widget w, Widget x, caddr_t call_data)
{
    
    /* This function will be called after the user has entered the 
       maximum and minimum values from the setRangeCallback. This
       function first calls CellGetVisibleRanges to get the current range.
       If the user leaves one field blank, the field will not be 
       effected. If the user changed some fields, this function  
       converts the text from char to double and passes the value 
       to the function CellSetVisibleRange; */
    
    char *t;   
    double tempDouble, minXLim, maxXLim, minYLim, maxYLim;
    char *str, *ptr;
    
    
    /* Create a warning message in case the user entered
       the max and min values as all zeros. */
    t = "Scale has been set to min(0,0) max(0,0).\n\
        Set Visible Range has failed.";
    
    CellGetVisibleRange(Cell, &minXLim, &minYLim, &maxXLim, &maxYLim);
    
    /* Get text from the 4 widgets and convert string from char to double. */
    str = XmTextGetString(Text[0]);
    tempDouble = strtod(str, &ptr);
    if (ptr != str)
        minXLim = tempDouble; 
    XtFree(str);
    
    str = XmTextGetString(Text[1]);
    tempDouble = strtod(str, &ptr); 
    if (ptr != str)
        minYLim = tempDouble; 
    XtFree(str);
    
    str = XmTextGetString(Text[2]);
    tempDouble = strtod(str, &ptr);
    if (ptr != str)
        maxXLim = tempDouble;  
    XtFree(str);
    
    str = XmTextGetString(Text[3]);
    tempDouble = strtod(str, &ptr);
    if (ptr != str)
        maxYLim = tempDouble;  
    XtFree(str);
   
    if (minXLim == 0.0 && maxXLim == 0.0 && minYLim == 0.0 && maxYLim == 0.0) 
        createWarningMessage(x, t);
    else { 
        XtUnmanageChild(x); 
        CellSetVisibleRange(Cell, minXLim, minYLim, maxXLim, maxYLim); 
    }
    
}

static void cancelCallback(Widget w, Widget x, caddr_t call_data)
{

    XtUnmanageChild(x);
	
}

static void getRangeCallback(Widget w, caddr_t client_data, caddr_t call_data)
{
    /* This function creates a message dialog that displays the coordinates
       that are returned from CellGetVisibleRange. */
       
    int n;
    double minXLim, maxXLim, minYLim, maxYLim;
    Widget infoBox;
    XmString xmstr, xmstr2;
    Arg wargs[5];
    char message[100];
   
    
    CellGetVisibleRange(Cell, &minXLim, &minYLim, &maxXLim, &maxYLim);           
   
    /* Create the string for display. */
    sprintf(message, "min (X,Y) = (%g , %g)\n\nmax (X,Y) = (%g , %g)",
        minXLim, minYLim, maxXLim, maxYLim);
    
    /* Create a messaage dialog to display coordinates. */
    infoBox = XmCreateMessageDialog(w, "infoBox", NULL, 0); 
    
    xmstr = XmStringCreateLtoR(message, XmSTRING_DEFAULT_CHARSET);
    xmstr2  = XmStringCreateSimple("Get Range Dialog");  
    n = 0;
    XtSetArg(wargs[n], XmNdialogTitle, xmstr2);   n++;
    XtSetArg(wargs[n], XmNmessageString, xmstr);  n++;   
    XtSetValues(infoBox, wargs, n);
       
      
    /* Unmanage the Cancel and Help widgets, since we won't need them. */ 
    XtUnmanageChild(XmMessageBoxGetChild (infoBox, XmDIALOG_CANCEL_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild (infoBox, XmDIALOG_HELP_BUTTON));
    
    XmStringFree(xmstr);
    XmStringFree(xmstr2);
    
    XtManageChild(infoBox);
}

static void zoomCallback(Widget w, XtPointer data, caddr_t call_data)
{
    /* This one callback is used for 3 zoom pushbuttons: Zoom In, Zoom Out,
       and Reset Zoom. */
       
    int i = (int ) data;
    
    switch ( i ) {
        case 1 : CellZoomIn(Cell); 
                 break;
        case 2 : CellZoomOut(Cell);
                 break;
        case 3 : CellResetZoom(Cell);
                 break;
    } 
    
}


static void createWarningMessage(Widget parent, char *text)
{

    /* This function creates a warning dialog that displays 
       the text. The text is received as a parameter. */
        
    Widget message;
    Arg wargs[5];
    int n;
    
    
    /* Create a warning dialog widget to display a message. */
    n = 0;
    XtSetArg(wargs[n], XmNmessageString,  XmStringCreateLtoR(text,
        XmSTRING_DEFAULT_CHARSET));  n++;    
    message = XmCreateWarningDialog(parent, "message", wargs, n);
    	
    /* Unmanage the Cancel and Help widgets, since we won't need them. */ 
    XtUnmanageChild(XmMessageBoxGetChild (message, XmDIALOG_CANCEL_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild (message, XmDIALOG_HELP_BUTTON));
    
    
    XtManageChild(message);
    
}    

static void printCallback(Widget w, caddr_t client_d, caddr_t call_d)
{
    /* This routine prints the Cell plot for the user.
       CellPrintContents writes a post script file that is passed to
       the utility routine PrintFile, a system independent routine
       queues the user for que information and queues the file 
       for printing. */
       
    static char fileName[] = "cellExamplePS.txt";
	
    CellPrintContents(Cell, fileName); 
#ifdef VMS
    PrintFile(w, fileName, "Cell_Output", True); 
#else
    PrintFile(w, fileName, "Cell_Output"); 
#endif /*VMS*/

}


static void exitCallback(Widget w, caddr_t client_data, caddr_t call_data)
{ 

    exit(0);

   
}


static int nFermiLogo(float x, float y)
{

    /* Simple logical function that draws the Fermi Logo.
       Returns 1 (TRUE) or 0 (FALSE) depending wether we are 
       on the dark (white) part of the Logo.
       The dimension of the Logo are approximately 1.0 by 1.0,
       leaving a bit of spacing for clarity. */
       
    int fermiLogo;
    float xa, ya, x1, y1, yMin, yMax, xx, yy, rOut2;
    static float barThickness = 0.2500,  lengthShort = 0.5417,
                posShort = 0.45,  lengthLong = 0.7500,
                posLong = 0.08333,  radiousIn = 0.166667,
                radiousOut = 0.083333,  xCircleOut = 0.416666,
                xCircleIn = 0.250,  topHeight = 0.80;
                
    /* Assume quadrant symmetry. */
    
    xa = fabs(x);
    ya = fabs(y);
    fermiLogo = 0;
    
    if ( xa > lengthLong )
        return(fermiLogo);
    if ( ya > topHeight)
        return(fermiLogo);
    if ( xa < lengthShort && (ya > posShort && ya <(posShort + barThickness))) {
        fermiLogo = 1;
        return(fermiLogo);
     }
    
        
    /* Assume eight fold symmetry. */
    
    if ( xa > ya ) {
        x1 = xa;
        y1 = ya;
     }
    else {
        x1 = ya;
        y1 = xa;
     }
     
    if (y1 < posLong)
        return(fermiLogo);
    if (x1 > xCircleOut && y1 > (posLong + barThickness))
        return(fermiLogo);
        
    /* More complicated case, do the circles... */
    
    if (x1 > xCircleOut)
    	yMax = posLong + barThickness;
    else if (x1 < (xCircleOut - radiousOut * 0.7071068))
    	    yMax = x1;
    else {
    	 xx = x1 - xCircleOut;
    	 rOut2 = radiousOut * radiousOut;
    	 yy = radiousOut - sqrt(rOut2 - xx * xx);
    	 yMax = posLong + barThickness + yy;
    }
    	 
    if (x1 > xCircleIn) 
    	yMin = posLong;
    else {
    	xx = xCircleIn - x1;
    	rOut2 = radiousIn * radiousIn;
    	yy = radiousIn - sqrt(rOut2 - xx * xx);
    	yMin = posLong + yy;
    }
    
    if (y1 < yMin || y1 > yMax)
    	return(fermiLogo);
    	
    fermiLogo = 1;
    return(fermiLogo);
}
